@@ -16,6 +16,7 @@
 #include <set>
 #include <algorithm>     // for all_of
 #include "asudata.hpp"   // for calculate_hkl_value_correlation
+#include "eig3.hpp"      // for eigen_decomposition
 #include "mtz.hpp"       // for Mtz
 #include "xds_ascii.hpp" // for XdsAscii
 #include "atox.hpp"      // for read_word
@@ -25,20 +26,6 @@
 
 namespace gemmi {
 
-inline bool parse_voigt_notation(const char* start, const char* end, SMat33<double>& b) {
-  bool first = true;
-  for (double* u : {&b.u11, &b.u22, &b.u33, &b.u23, &b.u13, &b.u12}) {
-    if (*start != (first ? '(' : ','))
-      return false;
-    first = false;
-    auto result = fast_from_chars(++start, end, *u);
-    if (result.ec != std::errc())
-      return false;
-    start = skip_blank(result.ptr);
-  }
-  return *start == ')';
-}
-
 class MtzToCif {
 public:
   enum Var { Dot=-1, Qmark=-2, Counter=-3, DatasetId=-4, Image=-5 };
@@ -57,6 +44,7 @@ public:
   double wavelength = NAN;           // user-specified wavelength
   int trim = 0;                      // output only reflections -N<=h,k,l<=N
   int free_flag_value = -1;          // -1 = auto: 0 or (if we have >50% of 0's) 1
+  std::string staraniso_version;     // for _software.version in "special_marker"
 
   static const char** default_spec(bool for_merged) {
     static const char* merged[] = {
@@ -99,41 +87,10 @@ public:
     return for_merged ? merged : unmerged;
   }
 
-  void write_cif(const Mtz& mtz, const Mtz* mtz2, std::ostream& os);
+  void write_cif(const Mtz& mtz, const Mtz* mtz2,
+                 SMat33<double>* staraniso_b, std::ostream& os);
   void write_cif_from_xds(const XdsAscii& xds, std::ostream& os);
 
-  void check_staraniso(const Mtz& mtz, std::ostream& out) {
-    size_t hlen = mtz.history.size();
-    for (size_t i = 0; i != hlen; ++i)
-      if (mtz.history[i].find("STARANISO") != std::string::npos) {
-        size_t version_pos = mtz.history[i].find("version:");
-        if (version_pos != std::string::npos)
-          staraniso_version = read_word(mtz.history[i].c_str() + version_pos + 8);
-        out << "History in merged MTZ includes scaling with STARANISO "
-            << staraniso_version << ".\n";
-        // StarAniso 2.3.74 (24-Apr-2021) and later write B tensor in history
-        bool got_b = false;
-        for (size_t j = i+1; j < std::min(i+4, hlen); ++j) {
-          const std::string& line = mtz.history[j];
-          if (starts_with(line, "B=(")) {
-            got_b = parse_voigt_notation(line.c_str() + 2,
-                                         line.c_str() + line.size(),
-                                         staraniso_b);
-            if (!got_b)
-              fail("failed to parse tensor Voigt notation: " + line);
-            break;
-          }
-        }
-        if (!got_b) {
-          out << "StarAniso B tensor not found. Intensities won't be checked.\n";
-          staraniso_b.u11 = NAN;
-        }
-        break;
-      }
-  }
-
-  const SMat33<double>& get_staraniso_b() const { return staraniso_b; }
-
 private:
   // describes which MTZ column is to be translated to what mmCIF column
   struct Trans {
@@ -146,11 +103,12 @@ private:
 
   // data corresponding to one sweep (dataset) in unmerged MTZ file
   struct SweepData {
-    int id;
-    int batch_count;
-    int offset;
-    const Mtz::Batch* first_batch;
-    const Mtz::Dataset* dataset;
+    int id = 0;
+    int batch_count = 0;
+    int offset = 0;
+    int crystal_id = 1;
+    const Mtz::Batch* first_batch = nullptr;
+    const Mtz::Dataset* dataset = nullptr;
   };
 
   std::vector<SweepData> sweeps;
@@ -158,9 +116,6 @@ private:
 
   std::vector<Trans> recipe;
 
-  SMat33<double> staraniso_b = {0., 0., 0., 0., 0., 0.};
-  std::string staraniso_version;
-
   const Trans* get_status_translation() const {
     for (const Trans& t: recipe)
       if (t.is_status)
@@ -168,10 +123,11 @@ private:
     return nullptr;
   }
 
-  void gather_sweep_data(const Mtz& mtz) {
+  bool gather_sweep_data(const Mtz& mtz) {
+    bool ok = true;
     int prev_number = INT_MIN;
     int prev_dataset = INT_MIN;
-    SweepData sweep{0, 0, 0, nullptr, nullptr};
+    SweepData sweep;
     for (const Mtz::Batch& batch : mtz.batches) {
       int dataset_id = batch.dataset_id();
       if (dataset_id != prev_dataset || batch.number != prev_number + 1) {
@@ -186,6 +142,7 @@ private:
           sweep.dataset = &mtz.dataset(dataset_id);
         } catch (std::runtime_error&) {
           sweep.dataset = nullptr;
+          ok = false;
           mtz.warn("Reference to absent dataset: " + std::to_string(dataset_id));
         }
       }
@@ -195,6 +152,7 @@ private:
     }
     if (sweep.id != 0)
       sweeps.push_back(sweep);
+    return ok;
   }
 
   // Get the first (non-zero) DWAVEL corresponding to an intensity, amplitude
@@ -441,8 +399,8 @@ inline bool validate_merged_mtz_deposition_columns(const Mtz& mtz, std::ostream&
 
 // note: both mi and ui get modified
 inline bool validate_merged_intensities(Intensities& mi, Intensities& ui,
-                                        const SMat33<double>& scale_aniso_b,
-                                        std::ostream& out) {
+                                        bool relaxed_check, std::ostream& out) {
+  // XDS files have 4 significant digits. Using accuracy 5x the precision.
   const double max_diff = 0.005;
   out << "Checking if both files match...\n";
   bool ok = true;
@@ -460,7 +418,14 @@ inline bool validate_merged_intensities(Intensities& mi, Intensities& ui,
       out << "(in the future, this app may recognize compatible space groups\n"
              "and reindex unmerged data if needed; for now, it's on you)\n";
   }
-  if (ui.unit_cell.approx(mi.unit_cell, 0.02)) {
+
+  auto eq = [](double x, double y, double rmsd) { return std::fabs(x - y) < rmsd + 0.02; };
+  if(eq(mi.unit_cell.a,     ui.unit_cell.a,     ui.unit_cell_rmsd[0]) &&
+     eq(mi.unit_cell.b,     ui.unit_cell.b,     ui.unit_cell_rmsd[1]) &&
+     eq(mi.unit_cell.c,     ui.unit_cell.c,     ui.unit_cell_rmsd[2]) &&
+     eq(mi.unit_cell.alpha, ui.unit_cell.alpha, ui.unit_cell_rmsd[3]) &&
+     eq(mi.unit_cell.beta,  ui.unit_cell.beta,  ui.unit_cell_rmsd[4]) &&
+     eq(mi.unit_cell.gamma, ui.unit_cell.gamma, ui.unit_cell_rmsd[5])) {
     out << "The same unit cell parameters.\n";
   } else {
     const UnitCell& mc = mi.unit_cell;
@@ -487,15 +452,10 @@ inline bool validate_merged_intensities(Intensities& mi, Intensities& ui,
   out << "Merged reflections: " << mi_size1 << ' ' << mi.type_str()
       << " (" << mi.data.size() << " w/o sysabs)\n";
 
-  bool relaxed_check = std::isnan(scale_aniso_b.u11);
-
-  if (!relaxed_check && !scale_aniso_b.all_zero()) {
+  if (mi.staraniso_b.ok()) {
     out << "Taking into account the anisotropy tensor that was used for scaling.\n";
-    for (Intensities::Refl& refl : ui.data) {
-      Vec3 hkl(refl.hkl[0], refl.hkl[1], refl.hkl[2]);
-      Vec3 s = ui.unit_cell.frac.mat.multiply(hkl);
-      refl.value *= std::exp(0.5 * scale_aniso_b.r_u_r(s));
-    }
+    for (Intensities::Refl& refl : ui.data)
+      refl.value *= mi.staraniso_b.scale(refl.hkl, ui.unit_cell);
   }
 
   // first pass - calculate CC and scale
@@ -513,47 +473,54 @@ inline bool validate_merged_intensities(Intensities& mi, Intensities& ui,
   int missing_count = 0;
   auto r1 = ui.data.begin();
   auto r2 = mi.data.begin();
+  auto refln_str = [](const Intensities::Refl& r) {
+    return r.intensity_label() + std::string(" ") + miller_str(r.hkl);
+  };
   while (r1 != ui.data.end() && r2 != mi.data.end()) {
-    if (r1->hkl == r2->hkl) {
-      double value1 = scale * r1->value;
-      double sigma1 = scale * r1->sigma; // is this approximately correct
-      double sq_value_max = std::max(sq(value1), sq(r2->value));
-      double sq_diff = sq(value1 - r2->value);
-      double weighted_sq_diff = sq_diff / (sq(sigma1) + sq(r2->sigma));
-      // XDS files have 4 significant digits. Using accuracy 5x the precision.
-      // Just in case, we ignore near-zero values.
-      if (!relaxed_check && sq_value_max > 1e-4 && sq_diff > sq(max_diff) * sq_value_max) {
-        if (differ_count == 0) {
-          out << "First difference: " << miller_str(r1->hkl)
-              << ' ' << value1 << " vs " << r2->value << '\n';
-        }
-        ++differ_count;
-        if (weighted_sq_diff > max_weighted_sq_diff) {
-          max_weighted_sq_diff = weighted_sq_diff;
-          max_diff_r1 = &*r1;
-          max_diff_r2 = &*r2;
+    if (r1->hkl == r2->hkl && r1->isign == r2->isign) {
+      if (!relaxed_check) {
+        double value1 = scale * r1->value;
+        double sigma1 = scale * r1->sigma; // is this approximately correct
+        double sq_max = std::max(sq(value1), sq(r2->value));
+        double sq_diff = sq(value1 - r2->value);
+        // Intensities may happen to be rounded to two decimal places,
+        // so if the absolute difference is <0.01 it's OK.
+        if (sq_diff > 1e-4 && sq_diff > sq(max_diff) * sq_max) {
+          if (differ_count == 0) {
+            out << "First difference: " << r1->hkl_label()
+                << ' ' << value1 << " vs " << r2->value << '\n';
+          }
+          ++differ_count;
+          double weighted_sq_diff = sq_diff / (sq(sigma1) + sq(r2->sigma));
+          if (weighted_sq_diff > max_weighted_sq_diff) {
+            max_weighted_sq_diff = weighted_sq_diff;
+            max_diff_r1 = &*r1;
+            max_diff_r2 = &*r2;
+          }
         }
       }
       ++r1;
       ++r2;
-    } else if (std::tie(r1->hkl[0], r1->hkl[1], r1->hkl[2]) <
-               std::tie(r2->hkl[0], r2->hkl[1], r2->hkl[2])) {
+    } else if (*r1 < *r2) {
       ++r1;
     } else {
       if (missing_count == 0)
-        out << "First missing reflection in unmerged data: " << miller_str(r1->hkl) << '\n';
+        out << "First missing reflection in unmerged data: " << refln_str(*r1) << '\n';
       ++missing_count;
       ++r2;
     }
   }
 
   if (differ_count != 0) {
-    const Miller& hkl = max_diff_r1->hkl;
-    out << "Most significant difference: " << miller_str(hkl) << ' '
+    out << "Most significant difference: " << refln_str(*max_diff_r1) << ' '
         << scale * max_diff_r1->value << " vs " << max_diff_r2->value << '\n';
     out << differ_count << " of " << corr.n << " intensities differ too much (by >"
         << to_str(max_diff * 100) << "%).\n";
-    ok = false;
+    if (differ_count >= 0.001 * corr.n)
+      ok = false;
+    else
+      out << "(less than 0.1% of all intensities -"
+          << " probably outlier rejection during merging)\n";
   }
   if (missing_count != 0) {
     out << missing_count << " out of " << mi.data.size()
@@ -567,13 +534,31 @@ inline bool validate_merged_intensities(Intensities& mi, Intensities& ui,
         out << " But other problems were found (see above).";
       out << '\n';
     } else {
-      out << "ERROR. Intensities do not match.\n";
+      out << (ok ? "OK. Intensities almost match.\n"
+                 : "ERROR. Intensities do not match.\n");
     }
   }
   return ok;
 }
 
-inline void MtzToCif::write_cif(const Mtz& mtz, const Mtz* mtz2, std::ostream& os) {
+#define WRITE(...) os.write(buf, gf_snprintf(buf, 255, __VA_ARGS__))
+
+inline void write_staraniso_b_in_mmcif(const SMat33<double>& b,
+                                       char* buf, std::ostream& os) {
+  double eigenvalues[3];
+  Mat33 eigenvectors = eigen_decomposition(b, eigenvalues);
+  const char* prefix = "\n_reflns.pdbx_aniso_B_tensor_eigen";
+  for (int i = 0; i < 3; ++i) {
+    double v = std::fabs(eigenvalues[i]) > 1e-4 ? eigenvalues[i] : 0;
+    WRITE("%svalue_%d %.5g", prefix, i+1, v);
+    for (int j = 0; j < 3; ++j)
+      WRITE("%svector_%d_ortho[%d] %.5g", prefix, i+1, j+1, eigenvectors[j][i]);
+  }
+  os << '\n';
+}
+
+inline void MtzToCif::write_cif(const Mtz& mtz, const Mtz* mtz2,
+                                SMat33<double>* staraniso_b, std::ostream& os) {
   if (mtz2 && mtz.is_merged() == mtz2->is_merged())
     fail("If two MTZ files are given, one must be merged and one unmerged,\n"
          "got two ", mtz.is_merged() ? "merged" : "unmerged");
@@ -581,7 +566,6 @@ inline void MtzToCif::write_cif(const Mtz& mtz, const Mtz* mtz2, std::ostream& o
   const Mtz* unmerged = mtz.is_merged() ? mtz2 : &mtz;
 
   char buf[256];
-#define WRITE(...) os.write(buf, gf_snprintf(buf, 255, __VA_ARGS__))
   if (with_comments) {
     os << "# Converted by gemmi-mtz2cif " GEMMI_VERSION "\n";
     for (const Mtz* m : {merged, unmerged})
@@ -601,42 +585,60 @@ inline void MtzToCif::write_cif(const Mtz& mtz, const Mtz* mtz2, std::ostream& o
   write_special_marker_if_requested(os, merged);
 
   if (unmerged) {
-    os << "_exptl_crystal.id 1\n\n";
-    bool scaled = (unmerged->column_with_label("SCALEUSED") != nullptr);
-    gather_sweep_data(*unmerged);
+    bool ok = gather_sweep_data(*unmerged);
+    std::set<const Mtz::Dataset*> used_datasets;
+    std::vector<std::string> crystal_names;
+    // Prepare used_datasets, crystal_names and set SweepData::crystal_id.
+    for (SweepData& sweep : sweeps)
+      if (sweep.dataset) {
+        used_datasets.insert(sweep.dataset);
+        if (ok) {
+          auto it = std::find(crystal_names.begin(), crystal_names.end(),
+                              sweep.dataset->crystal_name);
+          sweep.crystal_id = int(it - crystal_names.begin()) + 1;
+          if (it == crystal_names.end())
+            crystal_names.push_back(sweep.dataset->crystal_name);
+        }
+      }
 
-    os << "loop_\n_diffrn.id\n_diffrn.crystal_id\n_diffrn.details\n";
+    if (crystal_names.empty()) {
+      os << "_exptl_crystal.id 1\n";
+    } else {
+      os << "loop_\n_exptl_crystal.id\n";
+      for (size_t i = 0; i < crystal_names.size(); ++i)
+        os << i+1 << " # " << crystal_names[i] << '\n';
+    }
+
+    bool scaled = (unmerged->column_with_label("SCALEUSED") != nullptr);
+
+    os << "\nloop_\n"
+          "_diffrn.id\n_diffrn.crystal_id\n_diffrn.details\n";
     for (const SweepData& sweep : sweeps) {
-      os << sweep.id << " 1 '";
+      os << sweep.id << ' ' << sweep.crystal_id << " '";
       if (scaled)
         os << "scaled ";
       os << "unmerged data'\n";
     }
-    os << '\n';
 
-    os << "loop_\n"
+    os << "\nloop_\n"
           "_diffrn_measurement.diffrn_id\n"
           "_diffrn_measurement.details\n";
     for (const SweepData& sweep : sweeps)
       os << sweep.id << " '" << sweep.batch_count << " frames'\n";
-    os << '\n';
 
-    os << "loop_\n"
+    os << "\nloop_\n"
           "_diffrn_radiation.diffrn_id\n"
           "_diffrn_radiation.wavelength_id\n";
-    std::set<const Mtz::Dataset*> used_datasets;
     for (const SweepData& sweep : sweeps) {
       os << sweep.id << ' ';
-      if (sweep.dataset) {
-        used_datasets.insert(sweep.dataset);
-        os << sweep.dataset->id << '\n';
-      } else {
-        os << "?\n";
-      }
+      if (sweep.dataset)
+        os << sweep.dataset->id;
+      else
+        os << '?';
+      os << '\n';
     }
-    os << '\n';
 
-    os << "loop_\n"
+    os << "\nloop_\n"
           "_diffrn_radiation_wavelength.id\n"
           "_diffrn_radiation_wavelength.wavelength\n";
     for (const Mtz::Dataset* ds : used_datasets)
@@ -683,6 +685,9 @@ inline void MtzToCif::write_cif(const Mtz& mtz, const Mtz* mtz2, std::ostream& o
     write_cell_and_symmetry(cell, rmsds, mtz.spacegroup, buf, os);
   }
 
+  if (staraniso_b)
+    write_staraniso_b_in_mmcif(*staraniso_b, buf, os);
+
   if (merged)
     write_main_loop(*merged, buf, os);
   if (unmerged)
@@ -835,9 +840,9 @@ inline void MtzToCif::write_cif_from_xds(const XdsAscii& xds, std::ostream& os)
 
   write_special_marker_if_requested(os, false);
 
-  os << "_exptl_crystal.id 1\n\n";
+  os << "_exptl_crystal.id 1\n";
 
-  os << "loop_\n_diffrn.id\n_diffrn.crystal_id\n_diffrn.details\n";
+  os << "\nloop_\n_diffrn.id\n_diffrn.crystal_id\n_diffrn.details\n";
   for (const XdsAscii::Iset& iset : xds.isets)
     // We could write iset.input_info as details, but then local paths
     // could end up in the deposition.
@@ -930,22 +935,22 @@ inline void MtzToCif::write_cell_and_symmetry(const UnitCell& cell, double* rmsd
                                               const SpaceGroup* sg,
                                               char* buf, std::ostream& os) const {
   os << "_cell.entry_id " << entry_id << '\n';
-  WRITE("_cell.length_a    %8.3f\n", cell.a);
+  WRITE("_cell.length_a    %8.4f\n", cell.a);
   if (rmsds && rmsds[0] != 0.)
     WRITE("_cell.length_a_esd %7.3f\n", rmsds[0]);
-  WRITE("_cell.length_b    %8.3f\n", cell.b);
+  WRITE("_cell.length_b    %8.4f\n", cell.b);
   if (rmsds && rmsds[1] != 0.)
     WRITE("_cell.length_b_esd %7.3f\n", rmsds[1]);
-  WRITE("_cell.length_c    %8.3f\n", cell.c);
+  WRITE("_cell.length_c    %8.4f\n", cell.c);
   if (rmsds && rmsds[2] != 0.)
     WRITE("_cell.length_c_esd %7.3f\n", rmsds[2]);
-  WRITE("_cell.angle_alpha %8.3f\n", cell.alpha);
+  WRITE("_cell.angle_alpha %8.4f\n", cell.alpha);
   if (rmsds && rmsds[3] != 0.)
     WRITE("_cell.angle_alpha_esd %7.3f\n", rmsds[3]);
-  WRITE("_cell.angle_beta  %8.3f\n", cell.beta);
+  WRITE("_cell.angle_beta  %8.4f\n", cell.beta);
   if (rmsds && rmsds[4] != 0.)
     WRITE("_cell.angle_beta_esd %8.3f\n", rmsds[4]);
-  WRITE("_cell.angle_gamma %8.3f\n", cell.gamma);
+  WRITE("_cell.angle_gamma %8.4f\n", cell.gamma);
   if (rmsds && rmsds[5] != 0.)
     WRITE("_cell.angle_gamma_esd %7.3f\n", rmsds[5]);
   if (sg) {
